####
#### LD coefficient: D = p11 - p1 * p2
####
Delta = function(g1, g2){
  ## Esimate haplotype frequencies using EM algorithm
  m11 = sum(g1==0 & g2==0, na.rm=T)
  m12 = sum(g1==0 & g2==1, na.rm=T)
  m13 = sum(g1==0 & g2==2, na.rm=T)
  m21 = sum(g1==1 & g2==0, na.rm=T)
  m22 = sum(g1==1 & g2==1, na.rm=T)
  m23 = sum(g1==1 & g2==2, na.rm=T)
  m31 = sum(g1==2 & g2==0, na.rm=T)
  m32 = sum(g1==2 & g2==1, na.rm=T)
  m33 = sum(g1==2 & g2==2, na.rm=T)
  N = m11+m12+m13+m21+m22+m23+m31+m32+m33

  n11 = 2*m11 + m12 + m21
  n12 = m12 + 2*m13 + m23
  n21 = m21 + 2*m31 + m32
  n22 = m23 + m32 + 2*m33
  p11a = n11/(n11+n12+n21+n22)
  p12a = n12/(n11+n12+n21+n22)
  p21a = n21/(n11+n12+n21+n22)
  p22a = n22/(n11+n12+n21+n22)

  repeat {
    if(p11a*p22a+p12a*p21a==0){ptmp=0}
    else{ptmp = p11a*p22a / (p11a*p22a + p12a*p21a)}
    p11 = (n11+m22*ptmp)/(2*N)
    p12 = (n12+m22*(1-ptmp))/(2*N)
    p21 = (n21+m22*(1-ptmp))/(2*N)
    p22 = (n22+m22*ptmp)/(2*N)
    if(sum(abs(c(p11,p12,p21,p22)-c(p11a,p12a,p21a,p22a)))>1e-8){
      p11a=p11; p12a=p12; p21a=p21; p22a=p22
    } else {break}
   }

  p11 - (p11+p12)*(p11+p21)
}


####
#### Allele frequency calculation
####
allele.freq = function(g) {
  N = sum(!is.na(g))
  m1 = sum(g==1, na.rm=T)
  m2 = sum(g==2, na.rm=T)
  (m2 + .5*m1) / N
}


####
#### ATOM
####
## Descriptions of arguments:
##   pheno:      binary phenotype of interest
##   geno.test:  genotype data for analysis
##   geno.ref:   reference genotype data set
##   mkpos:      index positions of geno.test markers in the reference set.
##   cutoff:     a vector of cutoff values.  The default has five cutoff values:
##               0.8, 0.9, 0.95, 0.99, 0.999.
##   Nperm:      number of permutations.  The default value is 1000.
##
## Notes on the arguments:
##   The rows of 'pheno' and 'geno.test' must match to the same person.
##   If the marker names are known, 'mkpos' can be generated using
##     'match(rs.test, rs.ref)'.

ATOM = function(pheno, geno.test, geno.ref, mkpos,
                cutoff = c(.8,.9,.95,.99,.999), Nperm = 1000) {
  M = dim(geno.ref)[2]  ## #SNPs in reference data set
  n = length(pheno)  ## #subjects in real data

  ## Calculate pair-wise LD for the reference data
  D = diag(M)
  for(k in 1:M) {
    for(j in 1:k) {
      D[k,j] = D[j,k] = Delta(geno.ref[,k], geno.ref[,j])
    }
  }

  ## estimate allele frequencies using reference dataset
  p = numeric(M)
  for(j in 1:M) {
    p[j] = allele.freq(geno.ref[,j])
  }

  ## genotype score for all M SNPs
  geno.score = geno.test %*% diag(1/(p[mkpos]*(1-p[mkpos]))) %*% D[mkpos, ]

  ## PCA of the genotype scores
  PCA = prcomp(geno.score, scale=T)

  ## function to calculate minimun p over a range of cutoff values
  minpv = function(y) {
    ncutoff = length(cutoff)
    pv = numeric(ncutoff)
    for(i in 1:ncutoff) {
      npc = 1
      for(j in 1:n) {
        if(summary(PCA)$importance[3,j] > cutoff[i]) {
          npc=j;
          break;
        }
      }
      mod.PCA = summary(glm(y ~ PCA$x[,1:npc], family="binomial"))
      pv[i] = 1 - pchisq(mod.PCA$null.deviance - mod.PCA$deviance,
                         mod.PCA$df.null - mod.PCA$df.residual)
    }
    min(pv)
  }

  ## minimum p for real data
  pvalue.ATOM.min = minpv(pheno)

  ## permutations to evaluate significance
  pvalue.ATOM.min.perm = numeric(Nperm)
  for(i in 1:Nperm){
    pvalue.ATOM.min.perm[i] = minpv(sample(pheno))
  }

  ## permutation p-value
  sum(pvalue.ATOM.min.perm < pvalue.ATOM.min) / Nperm
}
